import json
import os
import re
import shutil
import subprocess
import tempfile
from pathlib import Path

from tala.cli import console_script
from tala.utils.chdir import chdir


class UnexpectedContentsException(Exception):
    pass


class TempDirTestMixin:
    def setup_class(cls):
        cls._get_installed_tala_version()
        if cls.no_tala_installed() or cls.tala_version_is_standard_release():
            cls._install_tala_as_editable()

    @classmethod
    def _get_installed_tala_version(cls):
        try:
            bytestring_response = subprocess.check_output("tala version", shell=True)
            cls._tala_version = bytestring_response.decode("utf-8")
        except subprocess.CalledProcessError:
            cls._tala_version = None

    @classmethod
    def no_tala_installed(cls):
        return cls._tala_version is None

    @classmethod
    def tala_version_is_editable(cls):
        return re.match(r"^\d+\.\d+\.\d+\.dev.*$", cls._tala_version)

    @classmethod
    def tala_version_is_standard_release(cls):
        if cls._tala_version is None:
            return False
        return re.match(r"^\d+\.\d+\.\d+$", cls._tala_version)

    @classmethod
    def _install_tala_as_editable(cls):
        subprocess.check_call("pip install -e .", shell=True)
        print("installed tala as editable")

    @classmethod
    def teardown_class(cls):
        if cls.tala_version_is_standard_release():
            cls._uninstall_tala()
            cls._install_old_tala_from_pip()

    def setup_method(self):
        self._temp_dir = tempfile.mkdtemp(prefix="TalaIntegrationTest")
        self._working_dir = os.getcwd()
        os.chdir(self._temp_dir)

    def teardown_method(self):
        os.chdir(self._working_dir)
        shutil.rmtree(self._temp_dir)

    def _uninstall_tala(self):
        subprocess.check_call("pip uninstall tala --yes", shell=True)

    def _install_old_tala_from_pip(self):
        subprocess.check_call(f"pip install tala=={self._tala_version}", shell=True)


class ConsoleScriptTestMixin(TempDirTestMixin):
    def setup_method(self):
        super().setup_method()
        self._process = None

    def _given_created_ddd_in_a_target_dir(self, name=None):
        self._create_ddd(name)

    def _create_ddd(self, name=None):
        name = name or "test_ddd"
        self._run_tala_with(["create-ddd", "--target-dir", "test_root", name])

    def _given_changed_directory_to_target_dir(self):
        return chdir("test_root")

    def _given_changed_directory_to_ddd_folder(self):
        return chdir("test_root/test_ddd")

    def _then_result_is_successful(self):
        def assert_no_error_occured():
            pass

        assert_no_error_occured()

    def _when_running_command(self, command_line):
        self._stdout, self._stderr = self._run_command(command_line)

    def _run_tala_with(self, args):
        console_script.main(args)

    def _run_command(self, command_line):
        self._process = subprocess.Popen(
            command_line.split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True
        )
        stdout, stderr = self._process.communicate()
        return stdout, stderr

    def _then_stderr_contains_constructive_error_message_for_missing_backend_config(self, config_path):
        pattern = "Expected backend config '.*{config}' to exist but it was not found. To create it, " \
                  r"run 'tala create-backend-config --filename .*{config}'\.".format(config=config_path)
        assert re.search(pattern, self._stderr) is not None

    def _then_stderr_contains_constructive_error_message_for_missing_ddd_config(self, config_path):
        pattern = "Expected DDD config '.*{config}' to exist but it was not found. To create it, " \
                  r"run 'tala create-ddd-config --filename .*{config}'\.".format(config=config_path)
        assert re.search(pattern, self._stderr) is not None

    def _given_config_overrides_missing_parent(self, path):
        self._set_in_config_file(path, "overrides", "missing_parent.json")

    def _set_in_config_file(self, path, key, value):
        with path.open(mode="r") as f:
            config = json.load(f)
        config[key] = value
        with path.open(mode="w") as f:
            string = json.dumps(config)
            f.write(str(string))

    def _then_file_contains(self, filename, expected_string):
        actual_content = self._read_file(filename)
        assert expected_string in actual_content

    def _read_file(self, filename):
        with open(filename) as f:
            actual_content = f.read()
        return actual_content

    def _then_stdout_contains(self, string):
        assert string in self._stdout, f"Expected {string} in stdout but got {self._stdout}"

    def _then_stderr_contains(self, string):
        assert string in self._stderr

    def _given_file_contains(self, filename, string):
        f = open(filename, "w")
        f.write(string)
        f.close()

    def _then_stdout_matches(self, expected_pattern):
        self._assert_matches(expected_pattern, self._stdout)

    def _then_stderr_matches(self, expected_pattern):
        self._assert_matches(expected_pattern, self._stderr)

    @staticmethod
    def _assert_matches(expected_pattern, string):
        assert re.search(
            expected_pattern, string
        ) is not None, f"Expected string to match '{expected_pattern}' but got '{string}'"

    def _given_ontology_contains(self, new_content):
        old_content = """
<ontology name="TestDddOntology">
</ontology>"""
        self._replace_in_file(Path("ontology.xml"), old_content, new_content)

    def _replace_in_file(self, path, old, new):
        with path.open() as f:
            old_contents = f.read()
        if old not in old_contents:
            raise UnexpectedContentsException(
                "Expected to find string to be replaced '{}' in '{}' but got '{}'".format(old, str(path), old_contents)
            )
        new_contents = old_contents.replace(old, new)
        with path.open("w") as f:
            f.write(new_contents)

    def _given_domain_contains(self, new_content):
        old_content = """
<domain name="TestDddDomain" is_super_domain="true">
  <goal type="perform" action="top">
    <plan>
      <forget_all/>
      <findout type="goal"/>
    </plan>
  </goal>
</domain>"""
        self._replace_in_file(Path("domain.xml"), old_content, new_content)

    def _given_rgl_is_disabled(self):
        config = Path("ddd.config.json")
        self._replace_in_file(config, '"use_rgl": true', '"use_rgl": false')

    def _given_rgl_is_enabled(self):
        config = Path("ddd.config.json")
        self._replace_in_file(config, '"use_rgl": false', '"use_rgl": true')



from tala.model.speaker import Speaker
from tala.model.move import MoveWithSemanticContent, Move, ICMMove, IssueICMMove, ICMMoveWithStringContent, \
    ICMMoveWithSemanticContent, ReportMove, PrereportMove, AnswerMove, RequestMove, AskMove


class MoveFactoryWithPredefinedBoilerplate(object):
    def __init__(
        self,
        ontology_name,
        understanding_confidence=None,
        speaker=None,
        utterance=None,
        ddd_name=None,
        perception_confidence=None
    ):
        self._ontology_name = ontology_name
        self._understanding_confidence = understanding_confidence
        self._perception_confidence = perception_confidence
        self._speaker = speaker
        self._utterance = utterance
        self._ddd_name = ddd_name

    def createMove(
        self,
        type_,
        content=None,
        understanding_confidence=None,
        speaker=None,
        utterance=None,
        modality=None,
        ddd_name=None,
        perception_confidence=None
    ):
        if understanding_confidence is None:
            understanding_confidence = self._understanding_confidence
        if perception_confidence is None:
            perception_confidence = self._perception_confidence
        if speaker is None:
            speaker = self._speaker
        if utterance is None:
            utterance = self._utterance
        if ddd_name is None:
            ddd_name = self._ddd_name

        kwargs = {
            "understanding_confidence": understanding_confidence,
            "speaker": speaker,
            "utterance": utterance,
            "modality": modality,
            "ddd_name": ddd_name,
            "perception_confidence": perception_confidence
        }

        if content is not None:
            classes = {
                Move.ANSWER: AnswerMove,
                Move.REQUEST: RequestMove,
                Move.ASK: AskMove,
            }
            if type_ in classes:
                Class = classes[type_]
                return Class(content, **kwargs)
            return MoveWithSemanticContent(type_, content, **kwargs)

        return Move(type_, **kwargs)

    def create_ask_move(self, question, speaker=None):
        return self.createMove(Move.ASK, question, speaker=speaker)

    def createAnswerMove(self, answer, speaker=None):
        return self.createMove(Move.ANSWER, answer, speaker=speaker)

    def createRequestMove(self, action):
        return self.createMove(Move.REQUEST, action, speaker=Speaker.USR, understanding_confidence=1.0)

    def createIcmMove(
        self,
        icm_type,
        content=None,
        content_speaker=None,
        polarity=None,
        understanding_confidence=None,
        speaker=None,
        ddd_name=None,
        perception_confidence=None
    ):
        if understanding_confidence is None:
            understanding_confidence = self._understanding_confidence
        if speaker is None:
            speaker = self._speaker

        if content is not None:
            if content == "issue":
                return IssueICMMove(
                    icm_type,
                    understanding_confidence=understanding_confidence,
                    speaker=speaker,
                    polarity=polarity,
                    ddd_name=ddd_name,
                    perception_confidence=perception_confidence
                )
            if isinstance(content, str):
                return ICMMoveWithStringContent(
                    icm_type,
                    content,
                    understanding_confidence=understanding_confidence,
                    speaker=speaker,
                    content_speaker=content_speaker,
                    polarity=polarity,
                    ddd_name=ddd_name,
                    perception_confidence=perception_confidence
                )
            return ICMMoveWithSemanticContent(
                icm_type,
                content,
                understanding_confidence=understanding_confidence,
                speaker=speaker,
                content_speaker=content_speaker,
                polarity=polarity,
                ddd_name=ddd_name,
                perception_confidence=perception_confidence
            )
        return ICMMove(
            icm_type,
            understanding_confidence=understanding_confidence,
            speaker=speaker,
            polarity=polarity,
            ddd_name=ddd_name,
            perception_confidence=perception_confidence
        )

    def create_report_move(self, report_proposition):
        return ReportMove(report_proposition)

    def create_prereport_move(self, service_action, arguments):
        return PrereportMove(self._ontology_name, service_action, arguments)



import fnmatch
import json
import re
import unittest
import warnings

import structlog

from tala.model.input_hypothesis import InputHypothesis
from tala.model.event_notification import EventNotification
from tala.utils.tdm_client import TDMClient, TDMRuntimeException
from tala.model.user_move import UserMove  # noqa: F401
from tala.model.interpretation import InterpretationWithoutUtterance, Interpretation  # noqa: F401
from tala.model.common import Modality


class EventNotReceivedException(Exception):
    pass


class InvalidConfidenceException(Exception):
    pass


class UnsupportedTurn(Exception):
    pass


class UnexpectedMoveFormatException(Exception):
    pass


class InteractionTestingTestCase(unittest.TestCase):
    def __init__(self, test, environment_or_url):
        self._test = test
        self._logger = structlog.get_logger(__name__)
        self._tdm_client = TDMClient(environment_or_url)

        method_name = self._ensure_name_can_be_used_as_python_method(test.name)
        self._create_test_method(method_name)
        unittest.TestCase.__init__(self, method_name)

        self.last_system_output_response = None
        self._detected_unexpected_passivity = False

    @property
    def last_system_output_response(self):
        return self._last_system_output_response

    @last_system_output_response.setter
    def last_system_output_response(self, new_response):
        def check_for_warnings():
            if new_response is not None and "warnings" in new_response and len(new_response["warnings"]) > 0:
                for warning in new_response["warnings"]:
                    warnings.warn(warning)

        check_for_warnings()
        self._last_system_output_response = new_response

    @property
    def name(self):
        return self._test.name

    @property
    def session(self):
        return self.last_system_output_response.get("session", None)

    @property
    def device_id(self):
        return {"device_id": "tala"}

    @staticmethod
    def _ensure_name_can_be_used_as_python_method(name):
        return re.sub(r"\W", "_", name)

    def _create_test_method(self, name):
        setattr(self, name, self.perform)

    def perform(self):
        self._logger.debug(f"Running interaction test '{self._test.name}' from file '{self._test.filename}'")
        self._start_session()
        for turn in self._test.turns:
            try:
                self._perform(turn)
            except TDMRuntimeException as exception:
                self._raise_and_log_assertion_error(
                    f"On line {turn.line_number} of {self._test.filename},\n\nencountered an error:\n  {exception}"
                )

    def _perform(self, turn):
        if turn.is_system_output_turn:
            self._handle_system_output_turn(turn)
        elif turn.is_user_input_turn:
            self._simulate_user_input(turn)
        elif turn.is_user_passivity_turn:
            self._simulate_user_passivity(turn)
        elif turn.is_event_turn:
            self._handle_event_notification(turn)
        else:
            raise UnsupportedTurn(f"Expected one of the supported turns but got '{turn}'.")

    def _start_session(self):
        self._update_last_system_output_response(self._tdm_client.start_session(session_data=self.device_id))

    def _update_last_system_output_response(self, response):
        file = open("tala-session.state", 'w')
        json.dump(response["session"], file)
        file.close()
        self.last_system_output_response = response

    def _handle_system_output_turn(self, turn):
        if turn.is_system_utterance_turn:
            self._expect_system_utterance(turn)
        elif turn.is_system_moves_turn:
            self._expect_system_moves(turn)
        else:
            raise UnsupportedTurn(f"Expected one of the supported system output turns but got '{turn}'.")

    def _simulate_user_input(self, turn):
        self._logger.debug("Simulating user input")
        if turn.is_recognition_hypotheses_turn:
            self._handle_recognition_hypotheses(turn)
        elif turn.is_user_interpretation_turn:
            self._handle_user_moves(turn)
        elif turn.is_user_semantic_input_turn:
            self._handle_user_interpretations(turn)
        else:
            raise UnsupportedTurn(f"Expected one of the supported user input turns but got '{turn}'.")

    def _handle_recognition_hypotheses(self, turn):
        self._update_last_system_output_response(self._send_natural_language_input_request(turn.hypotheses))

    def _handle_user_moves(self, turn):
        def interpretation(moves, utterance, modality):
            if utterance:
                modality = modality or Modality.SPEECH
                return Interpretation(moves, modality, utterance=utterance)
            modality = modality or Modality.OTHER
            return InterpretationWithoutUtterance(moves, modality)

        moves = [self._create_move(move_as_string) for move_as_string in turn.moves]
        interpretations = [interpretation(moves, turn.utterance, turn.modality)]
        self._update_last_system_output_response(self._tdm_client.request_semantic_input(interpretations, self.session))

    def _handle_user_interpretations(self, turn):
        body = {
            "session": self.session,
            "version": "3.4",
            "request": {
                "semantic_input": turn.semantic_input
            }
        }  # yapf: deactivate
        self._update_last_system_output_response(self._tdm_client.make_request(body))

    def _create_move(self, move):
        def is_semantic_expression(move):
            return isinstance(move, str)

        def is_move_object(move):
            return isinstance(move, dict)

        if is_move_object(move):
            return self._move_from_dict(move)
        if is_semantic_expression(move):
            return UserMove(move, perception_confidence=1.0, understanding_confidence=1.0)
        raise UnexpectedMoveFormatException(f"Expected either a semantic expression or a JSON object, but got {move}")

    def _move_from_dict(self, move):
        perception_confidence = move.get("perception_confidence", 1.0)
        understanding_confidence = move.get("understanding_confidence", 1.0)
        semantic_expression = move["semantic_expression"]
        return UserMove(
            semantic_expression,
            perception_confidence=perception_confidence,
            understanding_confidence=understanding_confidence
        )

    def _simulate_user_passivity(self, turn):
        self._logger.debug("Simulating user passivity")
        if self.last_system_output_response and \
           self.last_system_output_response["output"]["expected_passivity"] is not None:
            self._update_last_system_output_response(self._tdm_client.request_passivity(self.session))
        else:
            self._detected_unexpected_passivity = True
            self._line_number_of_unexpected_passivity = turn.line_number

    def _expect_system_utterance(self, system_utterance_turn):
        self._logger.debug(f"{self.__class__.__name__}._expect_system_utterance({str(system_utterance_turn)})")
        if self._detected_unexpected_passivity:
            self._raise_assertion_error_for_unexpected_passivity(system_utterance_turn)
        self._has_system_output_turn_been_processed = True
        actual_utterance = self._get_system_utterance()
        self._assert_system_utterance_matches(
            system_utterance_turn.utterance, actual_utterance, system_utterance_turn.line_number
        )

    def _expect_system_moves(self, system_moves_turn):
        self._logger.debug(f"{self.__class__.__name__}._expect_system_moves({str(system_moves_turn)})")
        if self._detected_unexpected_passivity:
            self._raise_assertion_error_for_unexpected_passivity(system_moves_turn)
        self._has_system_output_turn_been_processed = True
        actual_moves = self._get_system_moves()
        self._assert_system_moves_equals(system_moves_turn.moves, actual_moves, system_moves_turn.line_number)

    def _format_output(self, output):
        if isinstance(output, list):
            return json.dumps(output)
        return output

    def _raise_assertion_error_for_unexpected_passivity(self, system_turn):
        raise AssertionError(
            f"\n\nOn line {system_turn.line_number:d} of {self._test.filename},\n" +
            f"expected:\n  S> {self._format_output(system_turn.output)}\n\n" +
            f"in response to\n  U>\non line {self._line_number_of_unexpected_passivity:d}\n\n" +
            "but the system didn't expect user passivity."
        )

    def _raise_and_log_assertion_error(self, error_string):
        self._logger.debug(f"Assertion error: {error_string}")
        raise AssertionError(error_string)

    def _get_system_utterance(self):
        actual_utterance = self.last_system_output_response["output"]["utterance"]
        self._logger.debug(f"Got system utterance {actual_utterance!r}")
        return actual_utterance

    def _get_system_moves(self):
        actual_moves = self.last_system_output_response["output"]["moves"]
        self._logger.debug(f"Got system moves {actual_moves!r}")
        return actual_moves

    def _assert_system_utterance_matches(self, pattern, utterance, line_number):
        if not fnmatch.fnmatch(utterance, pattern):
            self._fail_with_mismatch(line_number, pattern, utterance)

    def _assert_system_moves_equals(self, expected_moves, actual_moves, line_number):
        if not expected_moves == actual_moves:
            self._fail_with_mismatch(line_number, expected_moves, actual_moves)

    def _fail_with_mismatch(self, line_number, expected, actual):
        self._raise_and_log_assertion_error(
            f"On line {line_number:d} of {self._test.filename},\n\n"
            f"expected:\n  S> {self._format_output(expected)}\n\n"
            f"but got:\n  S> {self._format_output(actual)}"
        )

    def _send_natural_language_input_request(self, hypotheses_tuples):
        hypotheses = [
            InputHypothesis(utterance, self._parse_score(score_as_string))
            for utterance, score_as_string in hypotheses_tuples
        ]
        return self._tdm_client.request_speech_input(hypotheses, self.session)

    def _parse_score(self, score):
        if score is None:
            return 1.0
        utterance_string_and_confidence_matcher = r"^\$(?P<level>\w+)|(?P<explicit>\d*\.\d+)?$"
        m = re.search(utterance_string_and_confidence_matcher, score)
        confidence_level_name = m.group("level")
        if confidence_level_name:
            raise NotImplementedError("Explicit names of confidence levels are not supported in this version.")
        explicit_score = m.group("explicit")
        if explicit_score:
            return float(explicit_score)
        raise InvalidConfidenceException(f"Invalid formatting of confidence: {score}")

    def _handle_event_notification(self, event_turn):
        if event_turn.is_notify_started_turn:
            self._update_last_system_output_response(
                self._tdm_client.request_event_notification(
                    EventNotification(event_turn.action, EventNotification.STARTED, event_turn.parameters), self.session
                )
            )
        else:
            raise UnsupportedTurn(f"Expected one of the supported event turns but got '{event_turn}'.")


class StringComparison(object):
    def __init__(self, string, pattern):
        self._string = string
        self._pattern = pattern
        self._match = self._string_matches_pattern(string, pattern)

    def match(self):
        return self._match

    def _string_matches_pattern(self, string, pattern):
        content = re.escape(pattern).replace(r'\*', '.*')
        re_pattern = f"^{content}$"
        return re.search(re_pattern, string, re.MULTILINE | re.DOTALL)

    def mismatch_description(self):
        return f"expected:\n---\n{self._mismatch_position_description_for_pattern()}\n---\n" \
               f"but got:\n---\n{self._mismatch_position_description_for_string()}"

    def _mismatch_position_description_for_pattern(self):
        position = self._get_mismatch_position()
        if position > 0:
            pattern_prefix = self._pattern[0:position]
            pattern_suffix = self._pattern[position:]
            return f"{pattern_prefix}======== HERE ========{pattern_suffix}"
        else:
            return self._pattern

    def _mismatch_position_description_for_string(self):
        position = self._get_mismatch_position()
        length = len(self._string)
        while length > 0:
            if self._string_matches_pattern(self._string[0:length], self._pattern[0:position]):
                string_prefix = self._string[0:length]
                string_suffix = self._string[length:]
                return f"{string_prefix}======== HERE? ========{string_suffix}"
            length = length - 1
        return self._string

    def _get_mismatch_position(self):
        length = len(self._pattern) - 1
        while length > 0:
            if self._string_matches_pattern(self._string, self._pattern[0:length] + "*"):
                return length
            length = length - 1
        return 0



import json
import logging
import os
import shutil
import tempfile

from mock import patch, ANY
import pytest
import structlog

from tala.log import logger


class TestLogger(object):
    def setup(self):
        self._log_filename = "mock_filename"
        self._temp_dir = tempfile.mkdtemp(prefix="LoggerTests")
        self._working_dir = os.getcwd()
        os.chdir(self._temp_dir)

    def teardown(self):
        os.chdir(self._working_dir)
        shutil.rmtree(self._temp_dir)

    @classmethod
    def tearDownClass(cls):
        structlog.reset_defaults()
        del logging.root.handlers[:]

    @patch("%s.TimeStamper" % logger.__name__)
    def test_configure_logger_configures_json_formatting(self, MockTimeStamper):
        self.given_timestamper_adds_timestamp(MockTimeStamper, "mock_timestamp")
        self.given_configured(logging.DEBUG)
        self.given_called_get_logger()
        self.when_invoke_debug("mock message", mock_key="mock_value")
        self.then_log_contains_json({
            "event": "mock message",
            "logger": "test_logger",
            "mock_key": "mock_value",
            "level": "debug",
            "time": "mock_timestamp"
        })

    def given_timestamper_adds_timestamp(self, MockTimeStamper, timestamp):
        def stamper(self, _, __, event_dict):
            event_dict["time"] = timestamp
            return event_dict

        MockTimeStamper.return_value = type("TimeStamper", (object, ), {"__call__": stamper})()

    def given_configured(self, level):
        logger.configure_file_logging(self._log_filename, level)

    def given_called_get_logger(self):
        self._logger = structlog.get_logger()

    def when_invoke_debug(self, message, **kwargs):
        self._logger.debug(message, **kwargs)

    def then_log_contains_json(self, expected_json):
        with open("logs/%s" % self._log_filename, "r") as log_file:
            actual_json = json.load(log_file)
        assert actual_json == expected_json

    @patch("%s.logging" % logger.__name__)
    def test_configure_file_logging_configures_file_handler(self, mock_logging):
        self.when_configuring_file_logging("mock_filename.log", "mock_log_level")
        self.then_stdlib_logging_is_configured_with_handlers(
            mock_logging, {
                "file": {
                    "class": "logging.handlers.WatchedFileHandler",
                    "filename": "logs/mock_filename.log",
                    "formatter": "plain",
                    "mode": "w",
                }
            }
        )

    def when_configuring_file_logging(self, filename, level):
        logger.configure_file_logging(filename, level)

    def then_stdlib_logging_is_configured_with_handlers(self, mock_logging, expected_handlers):
        actual_config = mock_logging.config.dictConfig.call_args[0][0]
        actual_handlers = actual_config["handlers"]
        assert actual_handlers == expected_handlers

    @patch("%s.logging" % logger.__name__)
    @pytest.mark.parametrize(
        "actual_level,expected_level", [
            (None, logging.WARNING),
            (logging.DEBUG, logging.DEBUG),
            (logging.INFO, logging.INFO),
            (logging.WARNING, logging.WARNING),
            (logging.ERROR, logging.ERROR),
            (logging.CRITICAL, logging.CRITICAL),
        ]
    )
    def test_configure_file_logging_configures_level(self, mock_logging, actual_level, expected_level):
        self.when_configuring_file_logging("mock_filename.log", actual_level)
        self.then_stdlib_logging_is_configured_with(
            mock_logging, {
                "version": 1,
                "formatters": ANY,
                "handlers": ANY,
                "loggers": {
                    "": {
                        "handlers": ANY,
                        "level": expected_level,
                        "propagate": False,
                    },
                }
            }
        )

    def then_stdlib_logging_is_configured_with(self, mock_logging, expected_config):
        actual_config = mock_logging.config.dictConfig.call_args[0][0]
        assert actual_config == expected_config

    @patch("%s.logging" % logger.__name__)
    @pytest.mark.parametrize(
        "actual_level,expected_level", [
            (None, logging.WARNING),
            (logging.DEBUG, logging.DEBUG),
            (logging.INFO, logging.INFO),
            (logging.WARNING, logging.WARNING),
            (logging.ERROR, logging.ERROR),
            (logging.CRITICAL, logging.CRITICAL),
        ]
    )
    def test_configure_stdout_logging_configures_level(self, mock_logging, actual_level, expected_level):
        self.when_configuring_stdout_logging(actual_level)
        self.then_stdlib_logging_is_configured_with(
            mock_logging, {
                "version": 1,
                "formatters": ANY,
                "handlers": ANY,
                "loggers": {
                    "": {
                        "handlers": ANY,
                        "level": expected_level,
                        "propagate": False,
                    },
                }
            }
        )

    @patch("%s.logging" % logger.__name__)
    def test_configure_stdout_logging_configures_stdout_handler(self, mock_logging):
        self.when_configuring_stdout_logging("mock_log_level")
        self.then_stdlib_logging_is_configured_with_handlers(
            mock_logging,
            {"stdout": {
                "class": "logging.StreamHandler",
                "stream": "ext://sys.stdout",
                "formatter": "plain",
            }}
        )

    def when_configuring_stdout_logging(self, level):
        logger.configure_stdout_logging(level)

    @patch("%s.JSONRenderer" % logger.__name__)
    @pytest.mark.parametrize(
        "configurer,args", [(logger.configure_file_logging, ["mock_filename"]), (logger.configure_stdout_logging, [])]
    )
    def test_configure_logger_configures_structlog_json_renderer(self, MockJSONRenderer, configurer, args):
        self.when_configuring(configurer, *args)
        self.then_structlog_is_configured_with_json_renderer(MockJSONRenderer, sort_keys=True)

    def when_configuring(self, configurer, *args):
        configurer(*args)

    def then_structlog_is_configured_with_json_renderer(self, JSONRenderer, **expected_kwargs):
        JSONRenderer.assert_called_once_with(**expected_kwargs)

    @patch("%s.TimeStamper" % logger.__name__)
    @pytest.mark.parametrize(
        "configurer,args", [(logger.configure_file_logging, ["mock_filename"]), (logger.configure_stdout_logging, [])]
    )
    def test_get_logger_configures_structlog_time_stamper(self, MockTimeStamper, configurer, args):
        self.when_configuring(configurer, *args)
        self.then_structlog_is_configured_with_timestamper(MockTimeStamper, fmt="iso", key="time")

    def then_structlog_is_configured_with_timestamper(self, TimeStamper, **expected_kwargs):
        TimeStamper.assert_called_once_with(**expected_kwargs)



import json

import requests
from requests.exceptions import RequestException

from tala.ddd.services.abstract_service_wrapper import AbstractServiceWrapper
from tala.model.service_invocation import PROTOCOL_VERSION
from tala.model.service_query_result import QueryResultFromService
from tala.model.service_action_outcome import SuccessfulServiceAction, FailedServiceAction

CURRENT_VERSION = "1.1"
DEPRECATED_PROTOCOL_VERSIONS = ["1.0"]
SUPPORTED_PROTOCOL_VERSIONS = DEPRECATED_PROTOCOL_VERSIONS + [CURRENT_VERSION]


class HttpServiceInvocationException(Exception):
    pass


class HttpServiceClient(AbstractServiceWrapper):
    def __init__(self, logger, endpoint, name, http_formatter):
        super(HttpServiceClient, self).__init__()
        self._logger = logger
        self._endpoint = endpoint
        self._name = name
        self._http_formatter = http_formatter
        self._session = requests.Session()

    def _filter_cache_by_request_type(self, response):
        request_type = json.loads(response.request.body)["request"]["type"]
        return True if request_type in ["query", "validator"] else False

    def _post(self, data, headers, session):
        def get_http_service_timeout(session):
            return float(session.get("dme_http_service_client_timeout", 2.0))

        http_service_timeout = get_http_service_timeout(session)
        self._logger.info(
            f"Calling service '{self._name}' on '{self._endpoint}' with headers {headers} and data {data} with timeout {http_service_timeout}."
        )
        try:
            response = self._session.post(
                self._endpoint, data=json.dumps(data), headers=headers, timeout=http_service_timeout
            )
            self._validate_response(response)
            return response
        except RequestException as exception:
            self._logger.exception("RequestException encountered.")
            raise HttpServiceInvocationException(str(exception))

    def _validate_response(self, response):
        def validate_http_status(response):
            response.raise_for_status()

        def validate_status(response):
            actual_status = response["status"]
            expected_statuses = ["success", "fail", "error"]
            if actual_status not in expected_statuses:
                raise HttpServiceInvocationException(
                    f"Expected 'status' to be one of {expected_statuses} but got {actual_status}"
                )
            if actual_status == "error":
                message = response["message"]
                raise HttpServiceInvocationException(
                    f"Expected service to succeed but it had an error with message: '{message}'."
                )

        def validate_version(response):
            actual_version = response["data"]["version"]
            if actual_version not in SUPPORTED_PROTOCOL_VERSIONS:
                raise HttpServiceInvocationException(
                    f"Expected one of the supported versions {SUPPORTED_PROTOCOL_VERSIONS} but got '{actual_version}'"
                )

        validate_http_status(response)
        payload = json.loads(response.text)
        validate_status(payload)
        validate_version(payload)

    def perform(self, action, parameters, session, context):
        def get_outcome_from_response(response):
            status = response["status"]
            if status == "success":
                return SuccessfulServiceAction()
            elif status == "fail":
                return FailedServiceAction(response["data"]["reason"])
            else:
                raise HttpServiceInvocationException(
                    "Expected 'success' or 'fail' as response status but got {status}."
                )

        response = self._post(
            data={
                "version": PROTOCOL_VERSION,
                "session": session,
                "request": {
                    "type": "request",
                    "name": action,
                    "parameters": self._parameter_bindings_to_json_dict(parameters, session)
                },
                "context": {
                    "active_ddd": context.active_ddd,
                    "facts": self._http_formatter.facts_to_json_object(context.facts, session),
                    "invocation_id": context.invocation_id
                }
            },
            headers={"Content-type": "application/json"},
            session=session
        )

        response = json.loads(response.text)
        return get_outcome_from_response(response)

    def query(self, question, parameters, min_results, max_results, session, context):
        response = self._post(
            data={
                "version": PROTOCOL_VERSION,
                "session": session,
                "request": {
                    "type": "query",
                    "name": question.get_predicate().get_name(),
                    "parameters": self._parameter_bindings_to_json_dict(parameters, session),
                    "min_results": min_results,
                    "max_results": max_results
                },
                "context": {
                    "active_ddd": context.active_ddd,
                    "facts": self._http_formatter.facts_to_json_object(context.facts, session),
                    "invocation_id": context.invocation_id
                }
            },
            headers={"Content-type": "application/json"},
            session=session
        )

        response = json.loads(response.text)
        self._assert_successful(response)
        result_dicts = response["data"]["result"]
        return [
            QueryResultFromService(result_dict["value"], result_dict["confidence"], result_dict["grammar_entry"])
            for result_dict in result_dicts
        ]

    def validate(self, validator_name, parameters, session, context):
        response = self._post(
            data={
                "version": PROTOCOL_VERSION,
                "session": session,
                "request": {
                    "type": "validator",
                    "name": self._name,
                    "parameters": self._parameter_bindings_to_json_dict(parameters, session)
                },
                "context": {
                    "active_ddd": context.active_ddd,
                    "facts": self._http_formatter.facts_to_json_object(context.facts, session),
                    "invocation_id": context.invocation_id
                }
            },
            headers={"Content-type": "application/json"},
            session=session
        )

        response = json.loads(response.text)
        self._assert_successful(response)
        return response["data"]["is_valid"]

    @staticmethod
    def _assert_successful(response):
        actual_status = response["status"]
        expected_status = "success"
        if not actual_status == expected_status:
            raise HttpServiceInvocationException(f"Expected status '{expected_status}' but got '{actual_status}'.")

    def _parameter_bindings_to_json_dict(self, bindings, session):
        def binding_to_json(binding):
            if binding.is_multiple_instance_binding:
                return [
                    self._http_formatter.fact_to_json_object(proposition, session)
                    for proposition in binding.propositions
                ]
            else:
                return self._http_formatter.fact_to_json_object(binding.proposition, session)

        return {binding.parameter.name: binding_to_json(binding) for binding in bindings}



from tala.config import BackendConfig


class EqualityAssertionTestCaseMixin:
    def assert_eq_returns_true_and_ne_returns_false_symmetrically(self, object1, object2):
        assert object1 == object2
        assert not (object1 != object2)
        assert object2 == object1
        assert not (object2 != object1)

    def assert_eq_returns_false_and_ne_returns_true_symmetrically(self, object1, object2):
        assert object1 != object2
        assert not (object1 == object2)
        assert object2 != object1
        assert not (object2 == object1)


def load_mockup_travel(component_set_loader):
    load_internal_ddds(component_set_loader, ["mockup_travel"], ".")


def load_internal_ddds(component_set_loader, ddds, package, rerank_amount=None):
    rerank_amount = rerank_amount or BackendConfig.DEFAULT_RERANK_AMOUNT
    component_set_loader.ensure_ddds_loaded(ddds, path=f"{package}/ddds", rerank_amount=rerank_amount)
